import dask.array
import google
import numpy as np
from abc import ABC, abstractmethod
import xarray as xr
import logging
from enstools.feature.identification._proto_gen import identification_pb2
from .subset_accessor import DescriptionAccessor


class IdentificationTechnique(ABC):
    """
    Base abstract class for new feature identification algorithms. Need to implement abstract methods precompute,
    identify and postprocess. An identification technique gets called for a dataset. The dataset is split along
    dimensions to parallelize the identification. The abstract method identify() is called for each spatial block of the
    dataset (2D, 3D), for example for each timestamp and for each ensemble member. Beforehand, precompute() is called
    once, here one-time precomputations can be done.
    """

    def __init__(self):
        self.pb_dataset = None

    @abstractmethod
    def precompute(self, dataset: xr.Dataset, **kwargs):
        """
        Abstract method, which is called before the identification. Here, precomputations can be done. If the user wants
        to add fields to the dataset (for example to output it later), it should be added here. The parallelized
        identify() needs to know the structure of the final dataset beforehand.

        Parameters
        ----------
        dataset : xarray.Dataset
                The dataset to process.

        Returns
        -------
        The dataset with additional initialized fields which will be populated in identify().
        """
        # self.precomputed_stuff = self.foo()...
        # dataset['bar'] = xr.zeros_like(dataset['foo'], dtype=int)
        # ...
        return dataset

    @abstractmethod
    def identify(self, process_data: xr.Dataset, object_block, **kwargs): # TODO update docstring
        """
        Abstract identification method. This gets called in parallel for each spatial block. # TODO what if procedure cant be parallelized?

        Parameters
        ----------
        process_data : xarray.Dataset
                The subset of data which is processed in parallel,
                datasets should contain same amount of dimensions as imput dataset, but parallelized ones have length 1
                dataarrays ins dataset are squeezed accordingly to 2D or 3D fields.
        object_block
                Protobuf reference to the block of objects where the feature descriptions have to go in to. One new
                feature can be added by calling the ObjectProperties.get_instance(pb2_type) API, see the templates.

        Returns
        -------
        process_data : xarray.Dataset
                The (altered) dataset
        """

        return process_data

    def identify_block(self, data_block: xr.Dataset, index_accessor, **kwargs):
        """
        Internal intermediate method which gets called for every data block and calls identify() on it. It also does
        some preparation and object result handling. It adds the identified objects into the protobuf struct.

        Parameters
        ----------
        data_block : xarray.Dataset
                The data block to be processed.
        index_accessor: dict
                dict of the parameters and their values which have been squeezed by parallelization (e.g. time)
        Returns
        -------
        The (altered) data block.
        """

        # called in parallel from map_blocks
        split_dims = kwargs.keys()
        split_string = ', '.join([str(dim) + ": " + str(data_block.coords[dim].data[0]) for dim in split_dims])
        print("Start processing data block with dimensions:   " + split_string)

        # get mapping which data fields' coords are getting squeezed (time, ens, ...). some fields may not have them all.
        squeeze_da_dim = []
        for split_dim in split_dims:
            for da_str in data_block.data_vars:
                if split_dim in data_block[da_str].coords:
                    squeeze_da_dim.append((da_str, split_dim))
                    data_block[da_str] = data_block[da_str].squeeze(dim=[split_dim])

        if index_accessor.data.ndim and index_accessor.data.size:
            object_block_ref = DescriptionAccessor.from_pb2root_to_object_list(self.pb_dataset, index_accessor.data)
        else:
            object_block_ref = DescriptionAccessor.from_pb2root_to_object_list(self.pb_dataset, None)

        data_block = self.identify(data_block, object_block_ref)

        for unsqueeze_da, unsqueeze_dim in squeeze_da_dim:
            data_block[unsqueeze_da] = data_block[unsqueeze_da].expand_dims(dim=unsqueeze_dim)

        # TODO geometry

        return data_block

    def execute(self, dataset: xr.Dataset, **kwargs):
        """
        Execute the identification of the features in the dataset. The properties of each feature are collected via the
        abstract identify() in parallel and here collected.

        Parameters
        ----------
        dataset : xarray.Dataset
                The dataset.

        Returns
        -------
        pb_dataset : identification_pb2.DatasetDescription
                The description struct containing all detected feature descriptions.
        ds_mapped : xarray.Dataset
                Dataset after being mapped via identify()
        """
        # build meta data for protobuf structure
        from ._proto_gen import identification_pb2
        self.pb_dataset = identification_pb2.DatasetDescription()

        # precomputations according to strategy - may alter the dataset
        prec_ds = self.precompute(dataset)
        if prec_ds is not None:
            dataset = prec_ds

        from enstools.misc import get_ensemble_dim, get_time_dim

        # dimension names along which data is split (time, ens, ...)
        split_dimensions = []

        ensemble_dim = get_ensemble_dim(dataset)
        ensemble_dim_in_input = (ensemble_dim is not None)
        if ensemble_dim_in_input and ensemble_dim in dataset.dims:
            split_dimensions.append(ensemble_dim)
        ensemble_dim_len = len(dataset.coords[ensemble_dim].values) if ensemble_dim_in_input else 1

        time_dim = get_time_dim(dataset)
        time_dim_in_input = (time_dim is not None)
        if time_dim_in_input and time_dim in dataset.dims:
            split_dimensions.append(time_dim)
        time_dim_len = len(dataset.coords[time_dim].values) if time_dim_in_input else 1
        # TODO init_time

        # make sure is sorted so indices work out at the end
        dataset = dataset.sortby(split_dimensions)

        # field of indices into descriptions
        index_field = xr.DataArray(coords=[dataset.coords[dim] for dim in split_dimensions]).astype(dtype=DescriptionAccessor)

        # rechunk along parallelized dimensions
        ds_rechunked = dataset.chunk(dict((dim, 1) for dim in split_dimensions))  # sizes: 1

        # init pb_dataset. set member + valid time for each element in lists if existent
        # also keep track of IDs and put them into index field, so we have a (data:ens,time...)->(ensid,timeid...) mapping
        # TODO init_time
        cur_indices = dict()
        for i in range(ensemble_dim_len):
            s_i = self.pb_dataset.sets.add()
            if ensemble_dim_in_input:
                s_i.member = dataset.coords[ensemble_dim].data[i]
                cur_indices[ensemble_dim] = s_i.member

            for t in range(time_dim_len):
                timeline_step = s_i.timeline.add()
                if time_dim_in_input:
                    valid_time = dataset.coords[time_dim].data[t]
                    timeline_step.valid_time = str(np.datetime_as_string(valid_time, unit='s'))
                    cur_indices[time_dim] = valid_time
                index_field.loc[cur_indices] = DescriptionAccessor(i if ensemble_dim_in_input else None, None, t if time_dim_in_input else None) # TODO init time

        # dict of values of the split dimensions -> to access them within each block
        dict_split_arrays = dict((dim, dataset.coords[dim]) for dim in split_dimensions)

        # map the blocks, computation here
        mb = xr.map_blocks(self.identify_block, ds_rechunked, args=[index_field], kwargs=dict_split_arrays, template=ds_rechunked)
        ds_mapped = mb.compute()  # num_workers=4

        # postprocess
        ds_mapped, self.pb_dataset = self.postprocess(ds_mapped, self.pb_dataset)

        return self.pb_dataset, ds_mapped

    @abstractmethod
    def postprocess(self, dataset: xr.Dataset, pb2_desc, **kwargs):
        """
        Abstract method for postprocessing operations.
        It takes as input the dataset and the corresponding descriptions, and returns them in an arbitrary altered
        state.

        Parameters
        ----------
        dataset : xarray.Dataset
                The dataset.
        pb2_desc : identification_pb2.DatasetDescription
                The description.
        kwargs

        Returns
        -------
        As inputs, but can be altered.
        """

        self.pb_dataset = None
        return dataset, pb2_desc
