      PROGRAM pv_streamer

      implicit none

c     -------------------------------------------------------------------------------
c     Declaration of variables
c     -------------------------------------------------------------------------------

c     Maximum length of contour arrays
      integer      maxcont
      parameter    (maxcont=100000)

c     Input contour
      character*200 inputfile
      character*200 inputfile_raw
      character*11 datestr
      integer      cn,sn
      real         c_lon(maxcont),c_lat(maxcont)
      real         s_lon(maxcont),s_lat(maxcont)

c     Resolution, left lowest coordinates and dimensions of the gridded array
      character*200 cdfname,varname
      real         dx,dy
      parameter    (dx=1.5,dy=1.5)
      real         xmin,ymin
      parameter    (xmin=-180.,ymin=-90)
      integer      nx,ny
      parameter    (nx=241,ny=121)  !(nx=720,ny=361)
      real         str2(nx,ny) 

c     Computational arrays
      real         x(maxcont),y(maxcont)
      integer      ifirst,ilast
      real         dist(maxcont)
      integer      label(maxcont)
      real         dmax

c     Search parameters
c     : min_length = Minimal accepted length for a streamer (km)
c     : max_length = Maximal accepted length for a streamer (km)
c     : close_enct = Close encounter distance (km)
c     : max_base   = Maximal length of streamer base line (km)
c     : base_angle = Minimal accepted angle between steamer base line and contour (deg)
c     : max_surgery= Maximal number of surgery steps (avoid endless loops)
c     : aratio     = Minimal ratio of streamer length and length of base line
c     : max_stream = If more than <max_stream> percent of the contour is part of a
c     :              streamer, the contour is classified as highly irregular and no streamer 
c     :              is accepted
      real         min_length
      parameter    (min_length=2000.)
      real         max_length
      parameter    (max_length=15000.)
      real         close_enct
      parameter    (close_enct=1500.)
      real         max_base
      parameter    (max_base=2000.)
      real         base_angle
      parameter    (base_angle=20.)
      integer      max_surgery
      parameter    (max_surgery=200.)
      real         aratio
      parameter    (aratio=2.)
      real         max_stream
      parameter    (max_stream=0.8)

c     Flag for output mode (1:2d list, 2: 1d ascii, 3:cdf)
      integer      output
      parameter    (output=3)

c     Outputfile
      character*200 outputfile
      character*200 outputfile_raw
      integer      nlabel
      real         area(maxcont),length(maxcont)
      real         clon(maxcont),clat(maxcont)

c     Auxiliary variables
      integer      i,j,k
      character    ch
      real         ldist,sdist
      integer      indm,indr
      integer      iswithin
      real         angle
      real         vx1,vy1,vx2,vy2
      integer      surgery,surgery1,surgery2,surgery3,surgery4
      integer      isurgery
      integer      flag
      integer      ind0
      integer      crefile
      integer      thetalevel
      real         lon,lat
      character*80 base
      integer      count
      integer      outlabel
      character*2  hem
      character*2  inp_hem
      integer      inp_level
      character*4  c_year(maxcont)
      character*2  c_month(maxcont),c_day(maxcont),c_hour(maxcont)
      character*2  c_hem(maxcont)
      integer      c_level(maxcont)
      real         c_thresh(maxcont)
      character*25 c_file(maxcont)

c     Externals
      real         sdis
      external     sdis

c     ---------------------------------------------------------------------------------
c     Preparations
c     ---------------------------------------------------------------------------------

c     Read in parameter file
      open(10,file='fort.10')
        read(10,'(a)') inputfile_raw
        read(10,'(a)') outputfile_raw
        read(10,*) inp_hem
        read(10,*) inp_level
      close(10)
      
      inputfile = trim(inputfile_raw)
      outputfile = trim(outputfile_raw)
      
      print*,inputfile,' -> ',outputfile,' ',inp_hem,inp_level

c     Read in contour data (both hemispheres, stack of isentropes).
      open(10,file=inputfile)
      cn=0
 100  continue
        cn=cn+1
        read(10,*,end=110)
     >               c_file(cn),
     >               c_hem(cn),c_level(cn),c_thresh(cn),   
     >               c_lon(cn),c_lat(cn)
        
        if ( (c_level(cn).ne.inp_level).or.(c_hem(cn).ne.inp_hem) ) then
           cn = cn -1
        endif
        goto 100
 110  continue
      close(10)
      cn=cn-1  
c      print*,c_file(0),' x ',c_hem(0)

c     Decide about the hemisphere
      if ( c_lat(1).gt.0. ) then
         hem = 'nh'
      else
         hem = 'sh'
      endif
      do i=2,cn
         if ( (c_lat(i).lt.0.).and.(hem.eq.'nh') ) then
            print*,'Inconsistent latitudes: must be all on NH or SH'
            stop
         endif
         if ( (c_lat(i).gt.0.).and.(hem.eq.'sh') ) then
            print*,'Inconsistent latitudes: must be all on NH or SH'
            stop
         endif
      enddo

c     Correct coordinates. Longitudes between -180 and +180, and latitudes as
c     positive values (distinction between northern and souther hemisphere is
c     made by the 'c_hem' array
      do i=1,cn
         if (c_lon(i).ge.180.) c_lon(i)=c_lon(i)-360.
         c_lat(i)=abs(c_lat(i))
      enddo

c     Init the output cdf file
      do i=1,nx
         do j=1,ny
            str2(i,j)=0.
         enddo
      enddo

c     Open output file
      open(50,file=outputfile,access='append')

c     ------------------------------------------------------------------------------------
c     Identify streamers
c     ------------------------------------------------------------------------------------

c        Extract contour on northern or southern hemisphere
         sn=0
         do i=1,cn
            sn=sn+1
            s_lon(sn)=c_lon(i)
            s_lat(sn)=c_lat(i)
         enddo

c        Force a closed contour
         sn=sn+1
         s_lon(sn)=s_lon(1)
         s_lat(sn)=s_lat(1)

c        Transform contour from lat/lon to x/y. This grid is used for fast check whether
c        the connecting line between two contour points is within or outside of the contour
         call latlon_to_xy (s_lat,s_lon,x,y,sn)

c        Get maximum step in new coordinates
         dmax=0.
         do i=1,sn-1
            ldist=sqrt((x(i+1)-x(i))**2+(y(i+1)-y(i))**2)
            if (ldist.gt.dmax) dmax=ldist
         enddo

c        Avoid nasty periodicity problems by extending list on left and right side
c        The interesting domain is found at indices sn+1...2*sn. To the left and right
c        the list is copied.
         do i=1,sn
            s_lon(i+sn)=s_lon(i)
            s_lat(i+sn)=s_lat(i)
            s_lon(i+2*sn)=s_lon(i)
            s_lat(i+2*sn)=s_lat(i)
            x(i+sn)=x(i)
            y(i+sn)=y(i)
            x(i+2*sn)=x(i)
            y(i+2*sn)=y(i)
         enddo
         ifirst=sn+1
         ilast=ifirst+sn-1

c        Init the label array
         do i=1,3*sn
            label(i)=0
         enddo

c        ----- Find close encounters  ------------------------------------------------

c        Get distances of data points along contour (relative to first point)
         call contourlength(s_lon,s_lat,dist,3*sn)

c        Find close encounters along the contour
         indm=ifirst
 200     continue
         
c           Move to the right and search for close encounter. If a close
c           encounter is found, its index is saved in <indr>. A close 
c           encounter is only accepted if the connecting line is completely
c           within the stratosphere
            indr=-1
            j=indm+1
 301        ldist=abs(dist(indm)-dist(j))
            if ((ldist.lt.min_length).and.(j.lt.3*sn)) then
               j=j+1
               goto 301
            else if ((ldist.lt.max_length).and.(j.lt.3*sn)) then
               sdist=sdis(s_lon(indm),s_lat(indm),s_lon(j),s_lat(j))
               if (sdist.lt.close_enct) then
                  iswithin=0
                  call check_line_inout (iswithin,
     >                                   x(indm),y(indm),x(j),y(j),
     >                                   x,y,sn,dmax)
                  if (iswithin.eq.1) indr=j
               endif
               j=j+1
               goto 301
            endif

c           Mark all points between the close encounters <indm...indr> with label 1. If no
c           streamer was found, go to the next event
            if (indr.ne.-1) then
               do i=indm,indr
                  label(i)=1.
               enddo
            else
               goto 600
            endif

c           ----- Surgery for identified streamer  -----------------------------------------

c           In the following section the base of the streamer is determined. This needs some
c           addition and removal of edge points. If any surgergy was needed, set a flag            
            isurgery=0
 400        continue
            surgery=0
            
c           Determine the angle between base line and contour at left edge (indm).  If the 
c           angle is too small, remove one edge point. Such a removal is only accepted 
c           if the new base line is still whithin the stratosphere.
            vx1=x(indr)-x(indm)
            vy1=y(indr)-y(indm)
            vx2=x(indm+1)-x(indm-1)
            vy2=y(indm+1)-y(indm-1)
            call getangle (vx1,vy1,vx2,vy2,angle)
            angle=abs(angle)
            if (angle.gt.base_angle) goto 401
            
            if (indm.lt.indr) then               
               iswithin=0
               call check_line_inout (
     >                              iswithin,
     >                              x(indm+1),y(indm+1),x(indr),y(indr),
     >                              x,y,sn,dmax)
               if (iswithin.eq.1) then 
                  surgery1=1
                  label(indm)=0
                  indm=indm+1
              endif
            endif

 401        continue

c           Determine the angle between base line and contour at right edge (indr). If the 
c           angle is too small, remove one edge point. Such a removal is only accepted 
c           if the new base line is still whithin the stratosphere. 
            vx1=x(indr)-x(indm)
            vy1=y(indr)-y(indm)
            vx2=x(indr+1)-x(indr-1)
            vy2=y(indr+1)-y(indr-1)
            call getangle (vx1,vy1,vx2,vy2,angle)
            angle=abs(angle)
            if (angle.gt.base_angle) goto 402

            if (indr.gt.indm) then               
               iswithin=0
               call check_line_inout (
     >                              iswithin,
     >                              x(indm),y(indm),x(indr-1),y(indr-1),
     >                              x,y,sn,dmax)
               if (iswithin.eq.1) then               
                  surgery2=1
                  label(indr)=0
                  indr=indr-1
              endif
            endif

 402        continue

c           Try to expand the streamer edge to the right. Such an expansion is accepted if
c           (a) the new length of the base length remains smaller than the threshold, (b) the
c           expansion does not lead to a merging with another streamer, (c) the resulting
c           edge angle remains larger than the threshold, and (d) the connecting base line
c           is completely within the stratosphere
            
            sdist=sdis(s_lon(indm),s_lat(indm),
     >                 s_lon(indr+1),s_lat(indr+1))
            if (sdist.gt.max_base) goto 403

            if ((label(indr+1).eq.1).or.(label(indr+2).eq.1)) goto 403

            vx1=x(indr+1)-x(indm)
            vy1=y(indr+1)-y(indm)
            vx2=x(indr+2)-x(indr)
            vy2=y(indr+2)-y(indr)
            call getangle (vx1,vy1,vx2,vy2,angle)
            angle=abs(angle)
            if (angle.lt.base_angle) goto 403

            iswithin=0
            call check_line_inout (
     >                           iswithin,
     >                           x(indm),y(indm),x(indr+1),y(indr+1),
     >                           x,y,sn,dmax)
            if (iswithin.eq.0) goto 403

            surgery3=1
            label(indr+1)=1
            indr=indr+1

 403        continue

c           Try to expand the streamer edge to the left. Such an expansion is accepted if
c           (a) the new length of the base length remains smaller than the threshold, (b) the
c           expansion does not lead to a merging with another streamer, (c) the resulting
c           edge angle remains larger than the threshold, and (d) the connecting base line
c           is completely within the stratosphere
            
            sdist=sdis(s_lon(indm-1),s_lat(indm-1),
     >                 s_lon(indr),s_lat(indr))
            if (sdist.gt.max_base) goto 404

            if ((label(indm-1).eq.1).or.(label(indm-2).eq.1)) goto 404

            vx1=x(indr)-x(indm-1)
            vy1=y(indr)-y(indm-1)
            vx2=x(indm+1)-x(indm-1)
            vy2=y(indm+1)-y(indm-1)
            call getangle (vx1,vy1,vx2,vy2,angle)
            angle=abs(angle)
            if (angle.lt.base_angle) goto 404

            iswithin=0
            call check_line_inout (
     >                           iswithin,
     >                           x(indm-1),y(indm-1),x(indr),y(indr),
     >                           x,y,sn,dmax)
            if (iswithin.eq.0) goto 404

            surgery4=1
            label(indm-1)=1
            indm=indm-1

 404        continue
            
c           Repeat the correction if a surgery step was performed. If the maximal number
c           of surgery steps is reached, no further correction is performed
            surgery=surgery1+surgery2+surgery3+surgery4
            if ((surgery.ne.0).and.(isurgery.lt.max_surgery)) then
               isurgery=isurgery+1
               goto 400
            endif

c           Check whether the identified streamer is accepet. If the streamer length is too 
c           small compared to its base line, it is not accepeted.
            ldist=dist(indr)-dist(indm)
            sdist=sdis(s_lon(indm),s_lat(indm),
     >                 s_lon(indr),s_lat(indr))
            if (ldist.lt.(aratio*sdist)) then
               do i=indm,indr
                  label(i)=0
               enddo
            endif

c        ----- End of surgery  -------------------------------------------------------
      
c        Check next data point. If a close encounter was found, the next point is
c        after the labeled points. This guarantees that nearby streamers do not overlap
 600     continue
         if (indr.ne.-1) then
            label(indr)=0
            label(indr+1)=0
            indm=indr+2
         else
            indm=indm+1
         endif
         if (indm.lt.ilast) goto 200

c        ----- Give each streamer a unique label  --------------------------------------
 
c        Handle periodicity (copy left and right parts back to central part)
         do i=1,sn
            if ((label(i).eq.1).and.(label(i+sn).ne.1)) then
               label(i+sn)=1
            else if ((label(i+2*sn).eq.1).and.(label(i+sn).ne.1)) then
               label(i+sn)=1
            endif
         enddo

c        Copy the central part to left and right
         do i=1,sn
            s_lon(i)=s_lon(i+sn)
            s_lat(i)=s_lat(i+sn)
            s_lon(i+2*sn)=s_lon(i+sn)
            s_lat(i+2*sn)=s_lat(i+sn)
            x(i)=x(i+sn)
            y(i)=y(i+sn)
            label(i)=label(i+sn)
            x(i+2*sn)=x(i+sn)
            y(i+2*sn)=y(i+sn)
            label(i+2*sn)=label(i+sn)
         enddo

c        Keep only the central part of the array x,y,label. This part becomes the 
c        new contour, which now gets indices 1..sn. Additionally, shift the indices
c        of the array so that the first index is not part of a streamer. This makes
c        the subsequent handling of the streamers easier. The same is true if the
c        last index is not a streamer point
         ind0=0
         do i=1,sn
            if (label(i).eq.0) then
               ind0=i
               exit
            endif
         enddo
         if (ind0.ne.0) then
            do i=ind0,ind0+sn-1
               s_lon(i-ind0+1)=s_lon(i)
               s_lat(i-ind0+1)=s_lat(i)
               x(i-ind0+1)=x(i)
               y(i-ind0+1)=y(i)
               label(i-ind0+1)=label(i)
            enddo
            label(sn)=0
         endif

c        Check whether this is a strange contour. If more than <max_stream> percent of the
c        contour is part of streamers, it is assumed that this is a highly irregular contour.
c        No streamers are accepted in this case.
         count=0
         do i=1,sn
            if (label(i).ne.0) count=count+1
         enddo
         if (real(count)/real(sn).gt.max_stream) then
            do i=1,sn
               label(i)=0
            enddo
         endif

c        Label the identified streamers (give each of them a unique integer label)
         nlabel=0
         flag=0
         do i=1,sn
            
            if ((label(i).eq.1).and.(flag.eq.0)) then
               nlabel=nlabel+1
               label(i)=nlabel
               flag=1
            else if ((label(i).eq.1).and.(flag.eq.1)) then
               label(i)=nlabel
            else if ((label(i).eq.0).and.(flag.eq.1)) then
               flag=0
            endif
            
         enddo

c        ---- Switch hemisphere if needed ----------------------------------------------

         if ( hem.eq.'sh' ) then
            do i=1,sn
               s_lat(i) = -s_lat(i)
            enddo
         endif

c        ----- Bring streamer to grid and write output  ---------------------------------

         if (output.eq.1) then
            
c           Identify grid points within streamer. The output is a gridded 2d array
            call streamer_to_grid (str2,xmin,ymin,dx,dy,nx,ny,
     >                             s_lon,s_lat,x,y,label,sn)

c           Write identified streamers to output
            do i=1,nx
               do j=1,ny
                  if (nint(str2(i,j)).ne.0) then
                     lon=xmin+real(i-1)*dx
                     if (lon.lt.0) lon=lon+360.
                     lat=ymin+real(j-1)*dy
                    write(50,'(a25,a3,i5,f10.2,2f10.2,a3)') 
     >                               c_file(1),inp_hem,
     >                               inp_level,c_thresh(1),
     >                               lon,lat,'st'
                  endif
               enddo
            enddo

         endif

c        ----- Write streamer as ascii output -------------------------------------------

         if (output.eq.2) then
            do i=1,sn
               if (label(i).ne.0) then
                  outlabel=1
               else
                  outlabel=0
               endif
               write(50,'(f8.2,f7.2,i2)')
     >               s_lon(i),s_lat(i),
     >               outlabel
            enddo
         endif

c        ---- Write 2D netCDF file ------------------------------------------------------

         if ( output.eq.3 ) then
         
c           Identify grid points within streamer. The output is a gridded 2d array
            call streamer_to_grid (str2,xmin,ymin,dx,dy,nx,ny,
     >                             s_lon,s_lat,x,y,label,sn)

            cdfname = outputfile
            varname = 'STREAMER'
            call writecdf2D(cdfname,
     >                     varname,str2,0.,
     >                      dx,dy,xmin,ymin,nx,ny,1,1)
         endif

c        Exit point for loop over different levels
 500     continue

c     -------------------------------------------------------------------
c     Final tasks
c     -------------------------------------------------------------------

c     Close output file
      close(50)

      end


c     ********************************************************************
c     * Streamer identification                                          *
c     ********************************************************************

c     -----------------------------------------------------------------------
c     Transform from lat/lon to x/y coordinates
c     -----------------------------------------------------------------------

      SUBROUTINE latlon_to_xy (lat,lon,x,y,npts)

c     Transform lat/lon data points into a x/y coordinate frame

      implicit none
      
c     Declaration of subroutine parameters
      integer npts
      real    lat(npts),lon(npts)
      real    x(npts),y(npts)

c     Mathematical constants
      real      pi
      parameter (pi=3.14159)

c     Auxiliary variables
      integer i
      
      do i=1,npts
         x(i)=(90.-lat(i))*sin(lon(i)*pi/180.)
         y(i)=(90.-lat(i))*cos(lon(i)*pi/180.)
      enddo

      end

c     -----------------------------------------------------------------------
c     Decide whether a connecting line is within or outside of contour
c     -----------------------------------------------------------------------

      SUBROUTINE check_line_inout (iswithin,
     >                             xpos1,ypos1,xpos2,ypos2,
     >                             x,y,npts,dmax)

c     A data line <xpos1,ypos1,xpos2,ypos2> is given. Check whether this line  lies within
c     or outside of the contour <x(npts),y(npts)>. If the line is within, set the
c     flag <iswithin=1>, otherwise the flag is <iswithin=0>. The contour must be given,
c     in x/y coordinates (not lat/lon). <dmax> ist the maximum distance between two data
c     points along the xy contour
      
      implicit none

c     Declaration of subroutine parameters
      integer  npts
      real     x(npts),y(npts)
      real     xpos1,ypos1,xpos2,ypos2
      integer  iswithin
      real     dmax

c     Set the number of sampling points between <xpos1,ypos1> and <xpos2,ypos2>
      integer   nsample
      parameter (nsample=20) 

c     Auxiliary variables
      integer  i
      real     xpos,ypos
      integer  isinstrat

c     Sample along the connecting line
      isinstrat=1
      do i=1,nsample
         xpos=xpos1+real(i)/real(nsample+1)*(xpos2-xpos1)
         ypos=ypos1+real(i)/real(nsample+1)*(ypos2-ypos1)
         call  check_point_inout (isinstrat,
     >                            xpos,ypos,
     >                            x,y,npts,dmax)
         if (isinstrat.eq.0) exit
      enddo             

c     Return flag
      iswithin=isinstrat

      end

c     -----------------------------------------------------------------------
c     Decide whether a data point is within or outside of contour
c     -----------------------------------------------------------------------

      SUBROUTINE check_point_inout (iswithin,
     >                              xpos,ypos,
     >                              x,y,npts,dmax)

c     A data point <xpos,ypos> is given. Check whether this data point lies within
c     or outside of the contour <x(npts),y(npts)>. If the point is within, set the
c     flag <iswithin=1>, otherwise the flag is <iswithin=0>. The contour must be given,
c     in x/y coordinates (not lat/lon). <dmax> is the maximum distance between data
c     points along the xy contour
      
      implicit none

c     Declaration of subroutine parameters
      integer  npts
      real     x(npts),y(npts)
      real     xpos,ypos
      integer  iswithin
      real     dmax

c     Region for detailed check
      real      distmin
      parameter (distmin=2.)
      
c     Auxiliary variables
      real      angle,incangle
      integer   k
      integer   step
      real      xdist,ydist,dist
      

c     Get the total angle around the contour
      angle=0.
      k=1
      step=1

 100  continue

c       Get the distance of the contour point from the sampling point      
        xdist=abs(x(k)-xpos)
        ydist=abs(y(k)-ypos)
        if (xdist.lt.ydist) then 
           dist=ydist
        else
           dist=xdist
        endif

c       Determine the step size from this distance
        if (dist.le.distmin) then
           step=1
        else
           step=nint((dist-distmin)/dmax)
           if (step.lt.1) step=1
        endif
        if ((k+step).ge.npts) then
           step=npts-k
        endif

c       Get the angle increment     
        call getangle(x(k)-xpos,y(k)-ypos,
     >                x(k+step)-xpos,y(k+step)-ypos,incangle)
        angle=angle+incangle

c       Increment the counter
        k=k+step
        if (k.lt.npts) goto 100

      
c     Decide whether this is an interior or exterior point
      if (abs(angle).lt.180.) then
         iswithin=0
      else
         iswithin=1
      endif

      end

c     -----------------------------------------------------------------------
c     Get the length of a contour line
c     -----------------------------------------------------------------------

      SUBROUTINE contourlength(x,y,xa,npts)

c     Given a contour line with coordinates <x(npts),y(npts)>, calculate the length 
c     <xa(npts)> of each point on the contour line

      implicit none

c     Declaration of subroutine parameters
      integer           npts
      real              x(npts)
      real              y(npts)
      real              xa(npts)

c     Auxiliary variables
      integer           i
      real              length
      real              lon1,lat1,lon2,lat2

c     External functions
      real              sdis
      external          sdis

      length=0.0
      xa(1)=0.0
      do i=1,npts-1
         lon1=x(i)
         lat1=y(i)
         lon2=x(i+1)
         lat2=y(i+1)
         length=length+sdis(lon1,lat1,lon2,lat2)
         xa(i+1)=length
      enddo

      END

c     ---------------------------------------------------------------------
c     Determine the angle between two vectors
c     ---------------------------------------------------------------------

      SUBROUTINE getangle (vx1,vy1,vx2,vy2,angle)

c     Given two vectors <vx1,vy1> and <vx2,vy2>, determine the angle (in deg)
c     between the two vectors.

      implicit none

c     Declaration of subroutine parameters
      real vx1,vy1
      real vx2,vy2
      real angle

c     Auxiliary variables and parameters
      real len1,len2,len3
      real val1,val2,val3
      real pi
      parameter (pi=3.14159265359)

      len1=sqrt(vx1*vx1+vy1*vy1)
      len2=sqrt(vx2*vx2+vy2*vy2)

      if ((len1.gt.0.).and.(len2.gt.0.)) then
         vx1=vx1/len1
         vy1=vy1/len1
         vx2=vx2/len2
         vy2=vy2/len2

         val1=vx1*vx2+vy1*vy2
         val2=-vy1*vx2+vx1*vy2

         len3=sqrt(val1*val1+val2*val2)

         if ( (val1.ge.0.).and.(val2.ge.0.) ) then
            val3=acos(val1/len3)
         else if ( (val1.lt.0.).and.(val2.ge.0.) ) then
            val3=pi-acos(abs(val1)/len3)
         else if ( (val1.ge.0.).and.(val2.le.0.) ) then
            val3=-acos(val1/len3)
         else
            val3=-pi+acos(abs(val1)/len3)
         endif
      else
         val3=0.
      endif

      angle=180./pi*val3

      END

c     -----------------------------------------------
c     Spherical distance
c     -----------------------------------------------

      real function sdis(xp,yp,xq,yq)

c
c     calculates spherical distance (in km) between two points given
c     by their spherical coordinates (xp,yp) and (xq,yq), respectively.
c
      real      re
      parameter (re=6370.)
      real      xp,yp,xq,yq,arg
      parameter (deg2rad=0.017453292519943)

      arg=sin(yp*deg2rad)*sin(yq*deg2rad)
      arg=arg+cos(yp*deg2rad)*cos(yq*deg2rad)*cos((xp-xq)*deg2rad)
      if (arg.lt.-1.) arg=-1.
      if (arg.gt.1.) arg=1.
      sdis=re*acos(arg)

      end

c     -----------------------------------------------
c     Get 2d gridded field of identified streamers
c     -----------------------------------------------

      subroutine streamer_to_grid (arr,xmin,ymin,dx,dy,nx,ny,
     >                             s_lon,s_lat,x,y,label,npts)

c     A contour with identified and labeled streamers is given in the
c     array <x(npts),y(npts),label(npts)>. The output array <arr(nx,yn)>
c     is a 0/label field which separates streamers from non-streamer
c     points. It is assumed that contour array starts with a non-streamer
c     points.
      
      implicit none
      
c     Declaration of subroutine parameters
      integer   nx,ny
      real      arr(nx,ny)
      real      xmin,ymin,dx,dy
      integer   npts
      real      x(npts),y(npts)
      integer   label(npts)
      real      s_lon(npts),s_lat(npts)

c     Numerical epsilon for real comparison
      real      eps
      parameter (eps=0.01)
      
c     Auxiliary variables
      integer     i,j,k
      real        latmin(nx),latmax(nx)
      real        dist
      real        lat,lon,xpos,ypos
      real        angle,incangle
      integer     flag,ind0      
      integer     indx
      integer     il,ir
      real        max
      integer     ilabel,nlabel,labelok(npts)
      character*2 hem
      real        tmp(nx,ny)
      integer     indy

c     Init the output array
      do i=1,nx
         do j=1,ny
            arr(i,j)=0.
         enddo
      enddo

c     Decide about the hemipshere and make all latitudes positive
      if ( s_lat(1).gt.0. ) then     
         hem='nh'
      else
         hem='sh'
      endif
      do k=1,npts
        s_lat(k) = abs(s_lat(k))
      enddo

c     Determine the northernmost and the southernmost points of the contour
      do i=1,nx
         latmin(i)=100.
         latmax(i)=-100.
      enddo
      do i=1,nx
        lon = xmin + real(i-1) * dx
	do k=1,npts
           if ( abs(s_lon(k)-lon).lt.dx ) then
	      if ( s_lat(k).lt.latmin(i) ) latmin(i) = s_lat(k)
	      if ( s_lat(k).gt.latmax(i) ) latmax(i) = s_lat(k)
           endif
        enddo
	do k=1,npts
           if ( abs(s_lon(k)-lon).lt.dx ) then
	      if ( label(k).ne.0 ) latmax(i) = 100.
           endif
        enddo
      enddo

c     Make the bands for tests a little bit broader
      do i=1,nx
         if (abs(latmin(i)-latmax(i)).lt.eps) then
            latmin(i)=latmin(i)-2.
            latmax(i)=latmax(i)+2.
         endif
      enddo
                      
c     Loop over all grid points
      do i=1,nx
         do j=1,ny

c           Check whether this point has a chance of being a part of a streamer
            lat=ymin+real(j-1)*dy
            lon=xmin+real(i-1)*dx
            if ((lat.lt.latmin(i)).or.(lat.gt.latmax(i))) goto 100

c           Transform from lat/lon to x/y coordinates
            call latlon_to_xy (lat,lon,xpos,ypos,1)

c           Get the total angle around the contour
            flag=0
            do k=1,npts
               
               if ((flag.eq.0).and.(label(k).ne.0)) then
                  ilabel=label(k)
                  flag=1
                  ind0=k
                  call getangle(x(k)-xpos,y(k)-ypos,
     >                          x(k+1)-xpos,y(k+1)-ypos,angle)

               else if ((flag.eq.1).and.(label(k).ne.0)) then
                  call getangle(x(k)-xpos,y(k)-ypos,
     >                          x(k+1)-xpos,y(k+1)-ypos,incangle)
                  angle=angle+incangle
                  
               else if ((flag.eq.1).and.(label(k).eq.0)) then
                  flag=0
                  angle=angle-incangle
                  call getangle(x(k-1)-xpos,y(k-1)-ypos,
     >                          x(ind0)-xpos,y(ind0)-ypos,incangle)
                  angle=angle+incangle                  

                  if (abs(angle).gt.180.) then
                     arr(i,j)=real(ilabel)
                  endif
               endif
                  
            enddo

c           Handle if last point is part of a streamer
            if ((flag.eq.1).and.(label(npts).eq.0)) then
               flag=0
               angle=angle-incangle
               call getangle(x(npts-1)-xpos,y(npts-1)-ypos,
     >                       x(ind0)-xpos,y(ind0)-ypos,incangle)
               angle=angle+incangle                  
               
               if (abs(angle).gt.180.) then
                  arr(i,j)=real(ilabel)
               endif
            endif

c           Exit point
 100        continue
            
         enddo
      enddo

c     Check that each streamer along the contour is also in the gridded array
      nlabel=0
      do i=1,npts
         if (label(i).gt.nlabel) nlabel=label(i)
      enddo
      do i=1,nlabel
         labelok(i)=0
      enddo
      do i=1,nx
         do j=1,ny
            ilabel=nint(arr(i,j))
            if (ilabel.ne.0) labelok(ilabel)=1
         enddo
      enddo
      do i=1,nlabel
         if (labelok(i).eq.0) then
            print*,'Error in gridding... Missing streamer'
c            stop
         endif
      enddo
      
c     Swap array if southern hemisphere
      if ( hem.eq.'sh' ) then
        do i=1,nx
	   do j=1,ny
	     tmp(i,j) = arr(i,j)
             arr(i,j) = 0.
	   enddo
        enddo
        do i=1,nx
	  do j=1,ny
            lat         = ymin + real(j-1) * dy
            indy        = nint( (-lat -ymin)/dy ) + 1
            arr(i,indy) = tmp(i,j)
          enddo
       enddo
      endif

      end           

c     -----------------------------------------------
c     Get 2d gridded field of contour
c     -----------------------------------------------

      subroutine contour_to_grid (arr,xmin,ymin,dx,dy,nx,ny,
     >                            lon,lat,npts)

c     The contour is given in <lon,lat,npts>. Bring the contour to the grid
c     array <arr(nx,ny)>, which is specifeid by <xmin,ymin,dx,dy,nx,ny>.

      implicit none
      
c     Declaration of subroutine parameters
      integer   nx,ny
      real      arr(nx,ny)
      real      xmin,ymin,dx,dy
      integer   npts
      real      lon(npts),lat(npts)

c     Auxiliary variables
      integer   i,j
      integer   indx,indy

c     Init output array
      do i=1,nx
         do j=1,ny
            arr(i,j)=0.
         enddo
      enddo

c     Simple gridding
      do i=1,npts
         indx=nint((lon(i)-xmin)/dx)+1
         indy=nint((lat(i)-ymin)/dy)+1
         if ((indx.ge.1).and.(indx.le.nx).and.
     >       (indy.ge.1).and.(indy.le.ny)) then
            arr(indx,indy)=1.
         endif
      enddo
   
      end


c     ********************************************************************
c     * Netcdf input/output                                              *
c     ********************************************************************

c     --------------------------------------------------------------------
c     Subroutines to write the netcdf output file
c     --------------------------------------------------------------------

      subroutine writecdf2D(cdfname,
     >                      varname,arr,time,
     >                      dx,dy,xmin,ymin,nx,ny,
     >                      crefile,crevar)

c     Create and write to the netcdf file <cdfname>. The variable
c     with name <varname> and with time <time> is written. The data
c     are in the two-dimensional array <arr>. The list <dx,dy,xmin,
c     ymin,nx,ny> specifies the output grid. The flags <crefile> and
c     <crevar> determine whether the file and/or the variable should
c     be created

      IMPLICIT NONE

c     Declaration of input parameters
      character*200 cdfname,varname
      integer nx,ny
      real arr(nx,ny)
      real dx,dy,xmin,ymin
      real time
      integer crefile,crevar

c     Further variables
      real varmin(4),varmax(4),stag(4)
      integer ierr,cdfid,ndim,vardim(4)
      character*200 cstname
      real mdv
      integer datar(14),stdate(5)
      integer i

C     Definitions
      varmin(1)=xmin
      varmin(2)=ymin
      varmin(3)=1050.
      varmax(1)=xmin+real(nx-1)*dx
      varmax(2)=ymin+real(ny-1)*dy
      varmax(3)=1050.
      cstname=trim(cdfname)//'_cst'
      ndim=4
      vardim(1)=nx
      vardim(2)=ny
      vardim(3)=1
      stag(1)=0.
      stag(2)=0.
      stag(3)=0.
      mdv=-999.98999

C     Create the file
      if (crefile.eq.0) then
         call cdfwopn(cdfname,cdfid,ierr)
         if (ierr.ne.0) goto 906
      else if (crefile.eq.1) then
         call crecdf(cdfname,cdfid,
     >        varmin,varmax,ndim,cstname,ierr)
         if (ierr.ne.0) goto 903

C        Write the constants file
         datar(1)=vardim(1)
         datar(2)=vardim(2)
         datar(3)=int(1000.*varmax(2))
         datar(4)=int(1000.*varmin(1))
         datar(5)=int(1000.*varmin(2))
         datar(6)=int(1000.*varmax(1))
         datar(7)=int(1000.*dx)
         datar(8)=int(1000.*dy)
         datar(9)=1
         datar(10)=1
         datar(11)=0            ! data version
         datar(12)=0            ! cstfile version
         datar(13)=0            ! longitude of pole
         datar(14)=90000        ! latitude of pole
         do i=1,5
            stdate(i)=0
         enddo
         call wricst(cstname,datar,0.,0.,0.,0.,stdate)
      endif

c     Write the data to the netcdf file, and close the file
      if (crevar.eq.1) then
         call putdef(cdfid,varname,ndim,mdv,
     >        vardim,varmin,varmax,stag,ierr)
         if (ierr.ne.0) goto 904
      endif
      call putdat(cdfid,varname,time,0,arr,ierr)
      if (ierr.ne.0) goto 905
      call clscdf(cdfid,ierr)

      return

c     Error handling
 903  print*,'*** Problem to create netcdf file ***'
      stop
 904  print*,'*** Problem to write definition ***'
      stop
 905  print*,'*** Problem to write data ***'
      stop
 906  print*,'*** Problem to open netcdf file ***'
      stop

      END
